/*
** Copyright (C) 1999 Christian Werner
** modified by InterRed GmbH 03/2003: getmatchword() renamed to getmatchwordbyprefix(); getmatchword() added;
** getfileinfoByBeginningOfWord() added, getIndexFilenum() modified
**
** This program and library is free software; you can redistribute it and/or
** modify it under the terms of the GNU (Library) General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU (Library) General Public License for more details.
**
** You should have received a copy of the GNU (Library) General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package swishej;

import java.io.*;
import java.util.*;

/**
 * Representation of SWISH-E index file
 */

class IndexFile {

    /**
     * File offset mapping.
     */
    static String indexchars = "abcdefghijklmnopqrstuvwxyz،ݟ&#;0123456789_\\|/-+=?!@$%^'\"`~,.<>[]{}";

    /**
     * SWISH-E file header.
     */
    static final String ifheader = "# SWISH format 1.3";

    /**
     * Line marking word stemming info.
     */
    static final String stheader = "# Stemming Applied: ";

    /**
     * Reference for reading index from file or URL.
     */
    Reader rd;

    /**
     * Table of file offsets.
     */
    int offsets[];

    /**
     * Highest short file number of index.
     */
    int maxfilenum;

    /**
     * Indicator for index' validity.
     */
    boolean valid;

    /**
     * Indicator for word stemming.
     */
    boolean stemming;

    /**
     * Table to remember file information.
     */
    FileHash fh;

    /**
     * Table of stop words.
     */
    StopHash sh;

    /**
     * Table of meta words.
     */
    MetaHash mh;

    /**
     * Table of file names.
     */
    Hashtable fn;

    /**
     * Index file header.
     */
    String header;

    /**
     * Own toLowerCase static method, does NOT convert non-ASCII
     * characters to lower case.
     * @param string string to be converted
     * @return string with all ASCII characters recoded to lower case
     */
    static String toLowerCase(String string) {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < string.length(); i++) {
	    int x = (int) string.charAt(i);
	    if (x >= 0 && x <= 128) {
	        sb.append(Character.toLowerCase(string.charAt(i)));
	    } else {
	        sb.append(string.charAt(i));
	    }
	}
	return sb.toString();
    }

    /**
     * Own toLowerCase static method, does NOT convert non-ASCII
     * characters to lower case.
     * @param ch character to be converted
     * @return character recoded to lower case if ASCII
     */
    static char toLowerCase(char ch) {
        int x = (int) ch;
	if (x >= 0 && x <= 128) {
	    return Character.toLowerCase(ch);
	}
	return ch;
    }

    /**
     * Creates and reads index file representation.
     * @param rd Reader for reading file or URL.
     * @exception IOException if file is invalid or unreadable
     */
    IndexFile(Reader rd) throws IOException {
        this.rd = rd;
	offsets = new int[Consts.MAXCHARS];
	maxfilenum = 0;
	valid = false;
	stemming = false;
	header = new String();
	fh = new FileHash();
	sh = new StopHash();
	mh = new MetaHash();
	fn = new Hashtable();
	readOffsets();
	if (!valid) {
	    throw new IOException("invalid index file");
	}
	getIndexFilenum();
	readStopwords();
	readFileOffsets();
	readMetaNames();
    }

    /**
     * Read offsets right after file header
     * @exception IOException on read error
     */
    void readOffsets() throws IOException {
        int i;
	byte b[] = new byte[1];
	char c = 0;
	boolean eof = false;
	for (i = 0; i < Consts.MAXCHARS; i++) {
	    offsets[i] = 0;
	}
	rd.seek(0);
	StringBuffer hd = new StringBuffer();
	int line = 0;
	while (!eof) {
	    eof = rd.read(b) != 1;
	    c = (char) (b[0] & 0xFF);
	    if (c == '#') {
	        StringBuffer sb = new StringBuffer();
	        do {
		    sb.append(c);
		    hd.append(c);
		    eof = rd.read(b) != 1;
		    c = (char) (b[0] & 0xFF);
		} while (!eof && c != '\n');
		hd.append(c);
		line++;
		if (line == 1) {
		    valid = sb.toString().startsWith(ifheader);
		} else if (sb.toString().startsWith(stheader)) {
		    try {
		        String s = sb.toString();
			s = s.substring(stheader.length());
			int flag = Integer.parseInt(s);
			stemming = flag != 0;
		    } catch (Exception e) {
		    }
		}
		continue;
	    }
            break;
	}
	header = hd.toString();
	i = 0;
	while (!eof && c != '\n') {
	    int k = Consts.MAXLONGLEN;
	    int num = 0;
	    while (!eof && Character.isDigit(c) && k > 0) {
	        num = (num * 10) + (c - '0');
		eof = rd.read(b) != 1;
		c = (char) (b[0] & 0xFF);
		k--;
	    }
	    offsets[i++] = num;
	}
    }

    /**
     * Determine number of files in index
     * @exception IOException on read error
     */
    void getIndexFilenum() throws IOException {
        int pos = offsets[Consts.FILELISTPOS];
	int len = offsets[Consts.FILEOFFSETPOS] - pos;
	rd.seek(pos);
	byte b[] = new byte[len];
	if (rd.read(b) != len) {
	    throw new IOException("premature eof");
	}
	for (int i = 0; i < len; i++) {
	    char c = (char) (b[i] & 0xFF);
	    if (c == '\n') {
	        maxfilenum++;
	    }
	}
	if (Consts.DEBUG) {
	    System.out.println("*** Number of indexed files: " + maxfilenum);
	}
        if (maxfilenum == 0)
          throw new IOException("empty index");
    }

    /**
     * Read stop words from index into stop word table
     * @exception IOException on read error
     */
    void readStopwords() throws IOException {
        StringBuffer word = new StringBuffer();
	byte b[] = new byte[1];
	rd.seek(offsets[Consts.STOPWORDPOS]);
	while (true) {
	    if (rd.read(b) != 1) {
	        break;
	    }
	    char c = (char) (b[0] & 0xFF);
	    if (c == '\n') {
	        break;
	    }
	    if (!Character.isSpace(c)) {
	        word.append(c);
	    } else {
	        sh.addstopword(word.toString());
		if (Consts.DEBUG) {
		    System.out.println("*** Add stop word: " +
				       word.toString());
		}
		word.setLength(0);
	    }
	}
    }

    /**
     * Read file long numbers (= offsets) from index into file table
     * @exception IOException on read error
     */
    void readFileOffsets() throws IOException {
	byte b[] = new byte[1];
	rd.seek(offsets[Consts.FILEOFFSETPOS]);
	int i = 0;
	if (rd.read(b) != 1) {
	    throw new IOException();
	}
	char c = (char) (b[0] & 0xFF);
	while (c != '\n') {
	    int k = Consts.MAXLONGLEN;
	    int num = 0;
	    while (Character.isDigit(c) && k-- > 0) {
	        num = (num * 10) + (c - '0');
		if (rd.read(b) != 1) {
		    throw new IOException();
		}
		c = (char) (b[0] & 0xFF);
	    }
	    if (Consts.DEBUG) {
	        System.out.println("*** Add file number: " + i + "->" + num);
	    }
	    fh.addfile(i++, num);
        }
    }

    /**
     * Read meta words from index
     * @exception IOException on read error
     */
    void readMetaNames() throws IOException {
	byte b[] = new byte[1];
	rd.seek(offsets[Consts.METANAMEPOS]);
	StringBuffer word = new StringBuffer();
	int metanum = 2;
	while (true) {
	    if (rd.read(b) != 1) {
	        break;
	    }
	    char c = (char) (b[0] & 0xFF);
	    if (c == '\n') {
	        break;
	    }
	    if (!Character.isSpace(c)) {
	        word.append(c);
		continue;
	    }
	    String s = new String(word);
	    int index = s.indexOf("\"");
	    if (index > 0) {
	        word.setLength(index - 1);
	    }
	    s = IndexFile.toLowerCase(word.toString());
	    mh.addmeta(s, metanum);
	    if (Consts.DEBUG) {
	        System.out.println("*** Add meta word: " + s);
	    }
	    word.setLength(0);
	    metanum++;
	    if (metanum % 128 == 0) {
	        metanum++;
	    }
	}
    }

    /**
     * Decode short file number from given result number.
     * @param num number from search result
     * @return short file number
     */
    int decodefilenum(int num) {
        int i, extra;
        for (i = 1, extra = 0; i < num; i++) {
            if ((i % 128) == 0) {
                extra++;
                i++;
            }
        }
        num -= extra;
        return num;
    }

    /**
     * Lookup file given result number and return file description
     * @param num number from search result
     * @return File information or null.
     */
    String lookupfile(int filenum) {
	int n = decodefilenum(filenum) - 1;
	n = fh.getfilenum(n);
	if (Consts.DEBUG) {
	    System.out.println("*** Lookup file num " + filenum);
	}
	String name = (String) fn.get(new Integer(n));
	if (name == null) {
	    try {
		rd.seek(n);
		byte b[] = new byte[1];
		StringBuffer sb = new StringBuffer();
		while (true) {
		    if (rd.read(b) != 1) {
		        break;
		    }
		    char c = (char) (b[0] & 0xFF);
		    if (c == '\n') {
		        break;
		    }
		    sb.append(c);
		}
		name = sb.toString();
		fn.put(new Integer(n), name);
	    } catch (IOException e) {
	        return null;
	    }
	}
	return name;
    }

    /**
     * Lookup word and position index pointer accordingly.
     * @param word word to be found
     * @return true if word found, false otherwise
     */
    boolean findWord(String word) {
        boolean found = false;
	char w0 = word.charAt(0);
	for (int i = 0; i < indexchars.length(); i++) {
	    if (w0 == indexchars.charAt(i)) {
	        int pos = offsets[i];
		if (pos <= 0 || pos >= offsets[Consts.METANAMEPOS]) {
		    break;
		}
	        try {
		    rd.seek(pos);
		} catch (IOException e) {
		    return false;
		}
		found = true;
		break;
	    }
	}
	if (Consts.DEBUG) {
	    System.out.println("*** Find word: " + word + ": " + found);
	}
        return found;
    }

    /**
     * Lookup word and make ResultList of it
     * @param word word to be found
     * @param meta meta number to be considered
     * @return new ResultList or null if not found
     */
    ResultList getfileinfo(String word, int meta) {
        ResultList rl = new ResultList();
	char c = 0;
	boolean eof = false;
	byte b[] = new byte[1];
	StringBuffer fword = new StringBuffer();
	if (Consts.DEBUG) {
	    System.out.println("*** Get file info: " + word + ", " + meta);
	}
	while (!eof) {
	    try {
	        eof = rd.read(b) != 1;
	    } catch (IOException e) {
	        eof = true;
		break;
	    }
	    if (eof) {
	        break;
	    }
	    c = (char) (b[0] & 0xFF);
	    if (c == ':') {
	        int comp = word.compareTo(fword.toString());
		if (comp == 0) {
		    break;
		}
		fword.setLength(0);
		if (comp > 0) {
		    while (!eof) {
		        try {
			    eof = rd.read(b) != 1;
			} catch (IOException e) {
			    eof = true;
			    break;
			}
			c = (char) (b[0] & 0xFF);
			if (c == 0) {
			    break;
			}
		    }
		    continue;
		}
		return rl;
	    }
	    fword.append(c);
	}
	if (eof || c == 0) {
	    return rl;
	}
	int countnum = 1;
	int filenum = 0;
	int rank = 0;
	int structure = 0;
	while (!eof && b[0] != 0) {
	    int x = 0;
	    do {
	        try {
		    eof = rd.read(b) != 1;
		} catch (IOException e) {
		    eof = true;
		    break;
		}
		if (b[0] == 0) {
		    return rl;
		}
		x *= 128;
		x += b[0] & 0x7F;
	    } while (!eof && (b[0] & 0x80) != 0);
	    if (eof) {
	        break;
	    }
	    if (x != 0) {
	        if (Consts.DEBUG) {
		    System.out.println("*** Get file info: cnt " +
				       countnum +  " val " + x);
		}
	        if (countnum == 1) {
		    filenum = x;
		    countnum++;
		} else if (countnum == 2) {
		    rank = x;
		    countnum++;
		} else if (countnum == 3) {
		    structure = x;
		    countnum++;
		} else if (countnum == 4) {
		    if (x == meta) {
		        rl.add(filenum, rank, structure);
		    }
		    countnum = 1;
                }
            }
	    try {
	        eof = rd.read(b) != 1;
	    } catch (IOException e) {
	        eof = true;
		break;
	    }
        }
	return rl.empty() ? null : rl;
    }


    /**
     * Lookup by beginning of a word and make ResultList of it
     * @param beginning beginning of words to be found
     * @param meta meta number to be considered
     * @return new ResultList or null if not found
     */
    ResultList getfileinfoByBeginningOfWord(String beginning, int meta) {
      ResultList rl = new ResultList();
      Hashtable list = new Hashtable();
      char c = 0;
      boolean eof = false;
      boolean wordFound = false;
      byte b[] = new byte[1];
      StringBuffer fword = new StringBuffer();
      if (Consts.DEBUG)
        System.out.println("*** Get file info (by beginning of word): " + beginning + ", " + meta);
      // ersten Buchstaben von beginning suchen und filepointer dahin bewegen
      if (!findWord(beginning)) {
          return null;
      }
    readfile:
      while (!eof) {
        wordFound = false;
        if (Consts.DEBUG)
          System.out.println(" searching word... ");
        while(!eof && !wordFound) {
          try {
            eof = rd.read(b) != 1;
          } catch (IOException e) {
            eof = true;
          }
          if (eof)
            break readfile;
          c = (char) (b[0] & 0xFF);
          // wort-beginn gematched?
          if (fword.length() > 0 && beginning.equals(fword.toString())) {
            wordFound = true;
            // filepointer bis zum ':' bewegen
            while(c != ':' && !eof)
            {
              try {
                eof = rd.read(b) != 1;
              } catch (IOException e) {
                eof = true;
              }
              if (eof)
                break readfile;
              c = (char) (b[0] & 0xFF);
              fword.append(c);
            }
            if (Consts.DEBUG)
              System.out.println("*** matched word: " + fword.toString());
            fword.setLength(0);
          }
          else if(fword.length() > 0 && fword.charAt(0)!=beginning.charAt(0)) {
            break readfile;
          }
          // wenn wortende, dann filepointer bis zum naechsten wortanfang bewegen
          else if(c == ':') {
            fword.setLength(0);
            while (!eof && c != 0) {
              try {
                eof = rd.read(b) != 1;
              } catch (IOException e) {
                eof = true;
              }
              if (eof)
                break readfile;
              c = (char) (b[0] & 0xFF);
            }
          }
          else
            fword.append(c);
        }
        if (Consts.DEBUG)
          System.out.println(" getting word-data... ");
        // suchwort also gefunden, nun daten lesen
        int countnum = 1;
        int filenum = 0;
        int rank = 0;
        int structure = 0;
        while (!eof && b[0] != 0) {
          int x = 0;

          do {
            try {
              eof = rd.read(b) != 1;
            } catch (IOException e) {
              eof = true;
            }
            x *= 128;
            x += b[0] & 0x7F;
          } while (!eof && (b[0] & 0x80) != 0);
          if (eof)
            break readfile;
          if(b[0] == 0)
            break;


          if (x != 0) {
            if (Consts.DEBUG) {
              System.out.println("*** Get file info: cnt " +
                                 countnum +  " val " + x);
            }
            if (countnum == 1) {
              filenum = x;
              countnum++;
            } else if (countnum == 2) {
              rank = x;
              countnum++;
            } else if (countnum == 3) {
              structure = x;
              countnum++;
            } else if (countnum == 4) {
              if (x == meta) {
                ResultList.mergeResult(filenum, rank, structure, list);
              }
              countnum = 1;
            }
          }
          try {
            eof = rd.read(b) != 1;
          } catch (IOException e) {
            eof = true;
            break;
          }
        }

      }
      if(!list.isEmpty()) {
        Enumeration e = list.elements();
        while (e.hasMoreElements()) {
          Result r = (Result) e.nextElement();
          rl.add(r.filenum, r.rank, r.structure);
        }
      }
      return rl.empty() ? null : rl;
    }


    /**
     * Match words against given prefix
     * @param prefix word prefix
     * @param first true for first match
     * @return word or null if not found
     */
    String getmatchwordbyprefix(String prefix, boolean first) {
        if (first && !findWord(prefix)) {
	    return null;
	}
	if (rd.tell() == offsets[Consts.STOPWORDPOS]) {
	    return null;
	}
	boolean eof = false;
	byte b[] = new byte[1];
	StringBuffer word = new StringBuffer();
	char c = 0;
	while (!eof && rd.tell() != offsets[Consts.STOPWORDPOS]) {
	    try {
	        eof = rd.read(b) != 1;
	    } catch (IOException e) {
	        eof = true;
		break;
	    }
	    if (eof) {
	        break;
	    }
	    c = (char) (b[0] & 0xFF);
	    if (c == ':') {
		while (!eof &&
		       rd.tell() != offsets[Consts.STOPWORDPOS]) {
		    try {
		        eof = rd.read(b) != 1;
		    } catch (IOException e) {
		        eof = true;
			break;
		    }
		    c = (char) (b[0] & 0xFF);
		    if (c == 0) {
		        break;
		    }
		}
		if (word.charAt(0) != prefix.charAt(0)) {
		    return null;
		}
		if (word.toString().indexOf(prefix) == 0) {
		    return word.toString();
		}
	        word.setLength(0);
	    } else {
	        word.append(c);
	    }
	}
	return null;
    }


    String getmatchword(String wordPart, boolean first, boolean matchEnding) {
        if(first) {
          try {
            rd.seek(offsets[0]);
          } catch (IOException e) {
            return null;
          }
        }
        if (rd.tell() == offsets[Consts.STOPWORDPOS]) {
            return null;
        }
        boolean eof = false;
        byte b[] = new byte[1];
        StringBuffer word = new StringBuffer();
        char c = 0;
        while (!eof && rd.tell() != offsets[Consts.STOPWORDPOS]) {
            try {
                eof = rd.read(b) != 1;
            } catch (IOException e) {
                eof = true;
                break;
            }
            if (eof) {
                break;
            }
            c = (char) (b[0] & 0xFF);
            if (c == ':') {
                while (!eof &&
                       rd.tell() != offsets[Consts.STOPWORDPOS]) {
                    try {
                        eof = rd.read(b) != 1;
                    } catch (IOException e) {
                        eof = true;
                        break;
                    }
                    c = (char) (b[0] & 0xFF);
                    if (c == 0) {
                        break;
                    }
                }
                if(matchEnding){
                  if (word.toString().indexOf(wordPart) >= 0 &&
                      word.toString().indexOf(wordPart) == word.toString().length()-wordPart.length()) {
                    return word.toString();
                  }
                }
                else {
                  if (word.toString().indexOf(wordPart) >= 0) {
                    return word.toString();
                  }
                }
                word.setLength(0);
            } else {
                word.append(c);
            }
        }
        return null;
    }


    /**
     * Dump keywords given initial into String array.
     * @param initial String giving the initial character
     * @return String[] containing words or null
     */
    String[] dumpWords(String initial) {
        if (!findWord(initial)) {
	    return null;
	}
	char firstc = initial.charAt(0);
        Vector v = new Vector();
	boolean eof = false;
	byte b[] = new byte[1];
	StringBuffer word = new StringBuffer();
	char c = 0;
	while (!eof) {
	    try {
	        eof = rd.read(b) != 1;
	    } catch (IOException e) {
	        eof = true;
		break;
	    }
	    c = (char) (b[0] & 0xFF);
	    if (c == ':') {
	        String w = word.toString();
	        if (w.charAt(0) == firstc) {
		    v.addElement(word.toString());
		}
		word.setLength(0);
		while (!eof) {
		    try {
		        eof = rd.read(b) != 1;
		    } catch (IOException e) {
		        eof = true;
			break;
		    }
		    c = (char) (b[0] & 0xFF);
		    if (c == 0) {
		        break;
		    }
		}
	        if (w.charAt(0) != firstc) {
		    break;
		}
		continue;
	    }
	    word.append(c);
	}
	if (v.size() == 0) {
	    return null;
	}
	String ret[] = new String[v.size()];
	for (int i = 0; i < v.size(); i++) {
	    ret[i] = (String) v.elementAt(i);
	}
	return ret;
    }
}
