/*
** Copyright (C) 1999-2001 Christian Werner
**
** This program and library is free software; you can redistribute it and/or
** modify it under the terms of the GNU (Library) General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU (Library) General Public License for more details.
**
** You should have received a copy of the GNU (Library) General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package swishej;

import java.io.*;
import java.net.*;

/**
 * Class providing the SWISH-E based keyword search.
 * @see swishej.Consts
 */

public class Search {
    /**
     * Perform keyword search.
     * @param words keywords to be searched
     * @param maxhits maximum number of hits
     * @param structure which document parts to be considered
     * @param index filename for loading the index file to be searched
     * @return search result
     */
    public static String search(String words, int maxhits, int structure,
				String index) {
	FReader in = null;
	IndexFile inf = null;
	try {
	    in = new FReader(index);
	} catch (IOException e) {
	    return "err: index file unreadable\n";
	}
	try {
	    inf = new IndexFile(in);
	} catch (IOException e) {
	    return "err: index file corrupt\n";
	}
	return SearchImpl.search(inf, words, maxhits, structure, null);
    }

    /**
     * Perform keyword search.
     * @param words keywords to be searched
     * @param maxhits maximum number of hits
     * @param structure which document parts to be considered
     * @param file file for loading the index file to be searched
     * @return search result
     */
    public static String search(String words, int maxhits, int structure,
				File file) {
	FReader in = null;
	IndexFile inf = null;
	try {
	    in = new FReader(file);
	} catch (IOException e) {
	    return "err: index file unreadable\n";
	}
	try {
	    inf = new IndexFile(in);
	} catch (IOException e) {
	    return "err: index file corrupt\n";
	}
	return SearchImpl.search(inf, words, maxhits, structure, null);
    }

    /**
     * Perform keyword search.
     * @param words keywords to be searched
     * @param maxhits maximum number of hits
     * @param structure which document parts to be considered
     * @param url URL for loading the index file to be searched
     * @param base base URL for documents in index or null
     * @return search result
     */
    public static String search(String words, int maxhits, int structure,
				URL url, URL base) {
	UReader in = null;
	IndexFile inf = null;
	try {
	    in = new UReader(url);
	} catch (IOException e) {
	    if (base != null) {
	        return
		    "<body>" +
		    "<h2>Error: index file unreadable.</h2>" +
		    "</body>";
	    }
	    return "err: index file unreadable\n";
	}
	try {
	    inf = new IndexFile(in);
	} catch (IOException e) {
	    if (base != null) {
	        return
		    "<body>" +
		    "<h2>Error: index file corrupt.</h2>" +
		    "</body>";
	    }
	    return "err: index file corrupt\n";
	}
	return SearchImpl.search(inf, words, maxhits, structure, base);
    }

    /**
     * Perform keyword search.
     * @param words keywords to be searched
     * @param maxhits maximum number of hits
     * @param structure which document parts to be considered
     * @param inf the loaded index file to be searched
     * @param base base URL for documents in index or null
     * @return search result
     */
    public static String search(String words, int maxhits, int structure,
				IndexFile inf, URL base) {
        if (inf == null) {
	    if (base != null) {
	        return
		    "<body>" +
		    "<h2>Error: no index file.</h2>" +
		    "</body>";
	    }
	    return "err: no index file\n";
	}
	return SearchImpl.search(inf, words, maxhits, structure, base);
    }

    /**
     * Load SWISH-E index file.
     * @param url URL of index file.
     * @return IndexFile or null if load was not successful
     */
    public static IndexFile loadindex(URL url) {
	UReader ur = null;
	IndexFile inf = null;
	try {
	    ur = new UReader(url);
	} catch (IOException e) {
	    System.out.println("loadindex:" + e);
	    return null;
	}
	try {
	    inf = new IndexFile(ur);
	} catch (IOException e) {
	    System.out.println("loadindex:" + e);
	    return null;
	}
	return inf;
    }
}
